/**
 * Module: giraffe
 * Version: 2.0.0
 * By: Tom Bremer <tom@tbremer.com>
 * URL: https://github.com/tbremer/Giraffe#readme
 */

(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global.Giraffe = factory());
}(this, (function () { 'use strict';

var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

function Node(_ref) {
  var id = _ref.id,
      label = _ref.label,
      data = _ref.data;

  if (!id && id !== 0) throw new Error("All Node's require an id");
  this.identity = id;
  this.properties = _extends({}, data);
  this.labels = label ? label.constructor === Array ? label : [label] : [];
  this.edges = [];
}

var shape = {
  identity: Number,
  properties: Object,
  labels: Array,
  edges: Array
};

var validConstructors = [String, Number];

function testConstructor(type) {
  return validConstructors.indexOf(type.constructor) > -1;
}

function Edge(_ref) {
  var from = _ref.from,
      through = _ref.through,
      id = _ref.id,
      data = _ref.data,
      label = _ref.label;

  if (!label || !testConstructor(label)) throw new Error('All Edges need a single Label');

  this.identity = id;
  this.from = from.identity;
  this.through = through.identity;
  this.label = label;
  this.properties = _extends({}, data);
}

var shape$1 = {
  identity: Number,
  properties: Object,
  label: String,
  from: Number,
  through: Number
};

function guid() {
  return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
    var r = Math.random() * 16 | 0;
    var v = c === 'x' ? r : r & 0x3 | 0x8;

    return v.toString(16);
  });
}

function checkProperties(node, properties) {
  var props = node.properties;

  /**
   * Property Checking
   * skip if key is `edge` since that is checked before.
   * check if key exists
   * check if key's value is the same as the node's value
   */

  for (var key in properties) {
    if (key === 'edges') continue;
    if (!(key in props)) return false;
    if (props[key] !== properties[key]) return false;
  }

  return true;
}

/**
 * Ensure data exists and
 * Ensure key is in data
 */
function lookForKey(key, data) {
  return Boolean(data && key in data);
}

/**
 * Loop through array of objects to check
 * all objects contain the correct keys and
 * all values have the same constructors
 */
function ensureObjectsShape(objects, shape) {
  if (!objects || objects.constructor !== Array) throw new Error('Objects needs to be an array');

  for (var idx in objects) {
    var object = objects[idx];

    for (var key in shape) {
      if (!(key in object) || object[key].constructor !== shape[key]) {
        throw new Error('Incorrect shape for ' + JSON.stringify(object));
      }
    }
  }

  return true;
}

/**
 * Build Edges when all Nodes are known
 * `this` is bound in it's callee
 */
function buildEdges(edges) {
  var built = [];

  for (var idx in edges) {
    var edge = edges[idx];
    var from = _extends({}, findById(edge.from, this.nodes)); // eslint-disable-line
    var through = _extends({}, findById(edge.through, this.nodes)); // eslint-disable-line

    built.push(_extends({}, edge, { from: from, through: through }));
  }

  return built;
}

/**
 * Search a group for an object containing the proper identity
 */
function findById(id, group) {
  for (var idx in group) {
    var node = group[idx];

    if ('identity' in node && node.identity === id) return node;
  }
}

/**
 * Given an ID and a group
 * return it's index
 * or -1
 * based on IDs
 */
function findIndexById(id, group) {
  for (var idx in group) {
    var node = group[idx];

    if ('identity' in node && node.identity === id) return idx;
  }

  return -1;
}

function Giraffe(dataset, callback) {
  if (dataset && dataset.constructor === Function && !callback) {
    callback = dataset;
    dataset = {};
  }

  Object.defineProperty(this, '_ids', { value: [] });
  Object.defineProperty(this, '_generateId', {
    value: function generateId() {
      var id = guid();

      if (this._ids.indexOf(id) !== -1) return this._generateId();

      this._ids.push(id);

      return id;
    }
  });

  this.nodes = [];
  this.edges = [];
  this.labels = {
    edges: {},
    nodes: {}
  };
  this.callback = callback;

  /**
   * for both Nodes and Edges
   * check if dataset exists and if key is within it
   * Ensure correct shape for each object
   */
  if (lookForKey('nodes', dataset) && ensureObjectsShape(dataset.nodes, shape)) this.nodes = dataset.nodes;
  if (lookForKey('edges', dataset) && ensureObjectsShape(dataset.edges, shape$1)) this.edges = dataset.edges;

  /**
   * Add id's to db._ids array
   * Build up this.labels object for Nodes
   */
  for (var idx in this.nodes) {
    var node = this.nodes[idx];
    this._ids.push(node.identity);
    if (!node.labels.length) continue;

    for (var _idx in node.labels) {
      var label = node.labels[_idx];

      if (!(label in this.labels.nodes)) this.labels.nodes[label] = [];
      this.labels.nodes[label].push(node.identity);
    }
  }

  /**
   * Add id's to db._ids array
   * Build up this.labels object for Edges
   */
  for (var _idx2 in this.edges) {
    var edge = this.edges[_idx2];
    this._ids.push(edge.identity);
    if (!edge.label) throw new Error('All Edges need a label');
    var _label = edge.label;

    if (!(_label in this.labels.edges)) this.labels.edges[_label] = [];
    this.labels.edges[_label].push(edge.identity);
  }
}

Giraffe.prototype.create = function create(label, data) {
  if (label.constructor === Object) {
    data = label;
    label = null;
  }

  var id = this._generateId();
  var node = new Node({ id: id, label: label, data: data });

  this.nodes.push(node);

  if (label) {
    var labelObj = this.labels.nodes;

    if (!(label in labelObj)) labelObj[label] = [];
    labelObj[label].push(node.identity);
  }

  if (this.callback) this.callback('create', node);

  return node;
};

Giraffe.prototype.update = function nodes(nodes, labels, data) {
  if (nodes.constructor !== Array) nodes = [nodes];
  if (labels.constructor === Object) {
    data = labels;
    labels = null;
  }
  if (labels !== null && labels.constructor !== Array) labels = [labels];
  if (!data) data = {};

  for (var idx in nodes) {
    var node = nodes[idx];
    var isEdge = 'label' in node;

    if (isEdge && labels) throw new TypeError('Edge Labels cannot be changed.');
    if (labels) node.labels = node.labels.concat(labels);

    node.properties = _extends({}, node.properties, data);
  }

  if (this.callback) this.callback('update', nodes);

  return nodes;
};

Giraffe.prototype.remove = function remove(nodes) {
  if (nodes.constructor !== Array) nodes = [nodes];

  for (var n in nodes) {
    var node = nodes[n];
    var identity = node.identity;

    /**
     * Removes all of this node's edges
     * Removes all label -> edge references.
     */

    for (var e in node.edges) {
      var edgeId = node.edges[e];
      var edge = findById(edgeId, this.edges);
      var label = edge.label,
          _identity = edge.identity;

      if (label in this.labels.edges) {
        var idx = this.labels.edges[label].indexOf(_identity);
        if (idx !== -1) this.labels.edges[label].splice(idx, 1);
      }

      /**
       * Remove the edge
       */
      this.edges.splice(findIndexById(edgeId, this.edges), 1);
    }

    /**
     * Remove this node from the labels properties.
     */
    for (var _idx3 in node.labels) {
      var label = node.labels[_idx3];
      if (!(label in this.labels.nodes)) continue; // label does not exist

      var nodeIdx = this.labels.nodes[label].indexOf(identity);

      if (nodeIdx === -1) continue; // node index does not exist in label

      this.labels.nodes[label].splice(nodeIdx, 1);
    }

    /**
     * remove all edges that reference this node
     */
    for (var _e in this.edges) {
      var _edge = this.edges[_e];
      if (!_edge || _edge.through !== identity) continue;

      var _node = this.nodes[_edge.from];
      if (!_node) continue;

      var _idx4 = _node.edges.indexOf(_edge.identity);
      if (_idx4 === -1) continue;

      /**
       * splice edge out of node's array
       */
      _node.edges.splice(_idx4, 1);

      /**
       * remove
       */
      this.edges[_e] = undefined;
    }

    /**
     * finally, remove node.
     */
    this.nodes.splice(findIndexById(identity, this.nodes), 1);
  }

  if (this.callback) this.callback('remove');
};

Giraffe.prototype.edge = function edge(from, through, label, data) {
  if (from.constructor !== Array) from = [from];
  if (through.constructor !== Array) through = [through];

  var edges = [];

  for (var f in from) {
    var _from = from[f];

    for (var t in through) {
      var _through = through[t];
      var id = this._generateId();
      var edg = new Edge({ id: id, label: label, data: data, from: _from, through: _through });
      var labelObj = this.labels.edges;

      if (!(label in labelObj)) labelObj[label] = [];
      labelObj[label].push(edg.identity);

      _from.edges.push(id);
      this.edges.push(edg);
      edges.push(edg);
    }
  }

  if (this.callback) this.callback('edge', buildEdges.call(this, edges));

  return edges;
};

Giraffe.prototype.query = function query(label, properties) {
  if (!label && !properties) label = properties = null;

  if (label && label.constructor === Object) {
    properties = label;
    label = null;
  }

  var results = [];

  for (var idx in this.nodes) {
    var node = this.nodes[idx];
    if (!node) continue;
    if (label && node.labels.indexOf(label) === -1) continue;

    var nodeContainsValidProps = checkProperties(node, properties);
    var edgeCheck = properties && 'edges' in properties;

    /**
     * Edge checking
     * get count of edges being checked
     * loop through all edges, checking `this.labels.edges` for the edge ids,
     * then checking the corresponding edge from property.
     * if edge count is less than edge's being checked move on.
     */
    if (edgeCheck) {
      var edgeTotal = properties.edges.length;
      var count = 0;

      for (var _idx5 in properties.edges) {
        var edgeName = properties.edges[_idx5];
        var internalNodeIdsFromEdge = [];

        if (!(edgeName in this.labels.edges)) continue; // edge is not known

        for (var idFromLabel in this.labels.edges[edgeName]) {
          var id = this.labels.edges[edgeName][idFromLabel];
          var edge = findById(id, this.edges);

          if (!edge) continue; //edges can be undefined from the `remove` method.
          internalNodeIdsFromEdge.push(edge.from);
        }

        if (internalNodeIdsFromEdge.indexOf(node.identity) > -1) count++; // edge exists and is found
      }

      if (count < edgeTotal) continue; // Node does not have the edge and we move on. through our parent loop.
    }

    if (properties && !nodeContainsValidProps) continue;

    /**
     * Build our node object without a prototype (now Hash Objects);
     * loop through all the edges and create a circular reference to the node
     * bind the node and it's relationship to the same return
     * This works in a single degree of separation so a node can have an edge
     * to itself without stack-overflowing.
     */
    var returnObj = _extends({}, node);

    for (var edgeIdx in returnObj.edges) {
      var edgeId = returnObj.edges[edgeIdx];
      var _edge2 = _extends({}, findById(edgeId, this.edges));
      var throughNode = findById(_edge2.through, this.nodes);

      _edge2.through = _extends({}, throughNode);
      _edge2.from = returnObj;

      returnObj.edges[edgeIdx] = _edge2;
    }

    results.push(returnObj);
  }

  if (this.callback) this.callback('query', results);

  return results;
};

return Giraffe;

})));
